/* Copyright (C) 2003-2004 Datapark corp. All rights reserved.
   Copyright (C) 2000-2002 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "dps_config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>

#include "dps_common.h"
#include "dps_utils.h"
#include "dps_id3.h"
#include "dps_vars.h"
#include "dps_textlist.h"

static int add_var(DPS_DOCUMENT *Doc, char *name,char *val){
	DPS_VAR		*Sec;
	
	if((Sec=DpsVarListFind(&Doc->Sections,name))){
		DPS_TEXTITEM	Item;
		bzero((void*)&Item, sizeof(Item));
		Item.section=Sec->section;
		Item.str=val;
		Item.section_name=name;
		DpsTextListAdd(&Doc->TextList,&Item);
	}
	return DPS_OK;
}

static int get_tag(DPS_DOCUMENT *Doc){
	char	year[5]="";
	char	album[31];
	char	artist[31];
	char	songname[31];
	char	name[31];
	char	*tag=Doc->Buf.buf+Doc->Buf.size-128;
	
	memcpy(songname,tag+3,30);
	songname[30]='\0';
	
	memcpy(artist,tag+33,30);
	artist[30]='\0';
	
	memcpy(album,tag+63,30);
	album[30]='\0';
	
	memcpy(year,tag+93,5);
	year[4]='\0';
	
	DpsRTrim(songname, " ");
	DpsRTrim(artist, " ");
	DpsRTrim(album, " ");
	
	strcpy(name,"MP3.Song");
	add_var(Doc,name,songname);
	
	strcpy(name,"MP3.Album");
	add_var(Doc,name,album);
	
	strcpy(name,"MP3.Artist");
	add_var(Doc,name,artist);
	
	strcpy(name,"MP3.Year");
	add_var(Doc,name,year);
	
	return DPS_OK;
}

/* 
	id3 header v2.3.0 http://www.id3.org/
	ID3	3
	version	1+1
    	flag	1
	size	4
*/

static char* get_id3(DPS_DOCUMENT *Doc){
	char	*ch;
	char	*buf_in=Doc->Buf.content;
	size_t	hdr_len=Doc->Buf.content-Doc->Buf.buf;
	size_t	cont_len=Doc->Buf.size-hdr_len;
	char	*artist=NULL;
	char	*album=NULL;
	char	*songname=NULL;
	char	name[64];
	int	tagcount=0;
	
	if(hdr_len>Doc->Buf.size)return NULL;
	
	ch = buf_in;
	
	if ( *(ch+6) == 'b'){
		/*
			extened header:
			size	4
			flag	2
			size of pagging 4
		*/
		ch +=20;
	}else{
		ch +=10;
	}
	
	while(1){
		/*
			frame header:
			frame id 4
			size	 4
			flags	 2 
		*/
		
		unsigned char	frame_size = (unsigned char)*(ch+7);
		size_t		len = frame_size>cont_len?cont_len:frame_size;
		
		if (!strncmp(ch , "TPE1", 4)){
			ch +=10;
			artist = DpsMalloc(len+1);
			if (artist == NULL) break;
			dps_snprintf(artist, len, "%s", ch+1);
			artist[len]='\0';
			DpsRTrim(artist," ");
			if (++tagcount == 3)
					break;
		}else if(!strncmp(ch , "TALB", 4)){
			ch +=10;
			album = DpsMalloc(len+1);
			if (album == NULL) break;
			dps_snprintf(album, len, "%s", ch+1);
			album[len]='\0';
			DpsRTrim(album," ");
			if (++tagcount == 3)
				break;
		}else if(!strncmp(ch , "TIT2", 4) ){
			ch +=10;
			songname = DpsMalloc(len+1);
			if (songname == NULL) break;
			dps_snprintf(songname, len, "%s", ch+1);
			songname[len]='\0';
			DpsRTrim(songname," ");
			if (++tagcount == 3)
				break;
		}else if (((size_t)(ch - buf_in)+frame_size) <cont_len){
			ch +=10;
		}else{
			break;
		}
		ch +=frame_size;
	}
	
	if (!artist) artist = (char*)DpsStrdup("");
	if (!album) album = (char*)DpsStrdup("");
	if (!songname) songname = (char*)DpsStrdup("");
	
	strcpy(name,"MP3.Song");
	add_var(Doc,name,songname);
	
	strcpy(name,"MP3.Album");
	add_var(Doc,name,album);
	
	strcpy(name,"MP3.Artist");
	add_var(Doc,name,artist);
	
	DPS_FREE(artist);
	DPS_FREE(album);
	DPS_FREE(songname);
	
	return DPS_OK;
}

int DpsMP3Type(DPS_DOCUMENT *Doc){
	int	hd=((unsigned char)(Doc->Buf.content[0])+256*(unsigned char)(Doc->Buf.content[1])) & 0xf0ff;
	
	if(hd == 0xf0ff)
		return DPS_MP3_TAG;
        
        if (!strncmp(Doc->Buf.content,"RIFF",4))
		return DPS_MP3_RIFF;
	
        if (!strncmp(Doc->Buf.content, "ID3", 3))
		return DPS_MP3_ID3;
	
	return DPS_MP3_UNKNOWN;
}

int DpsMP3Parse(DPS_AGENT *A,DPS_DOCUMENT *Doc){
	if (!strncmp(Doc->Buf.content, "ID3", 3))
		get_id3(Doc);
	
	if ((Doc->Buf.size>=128) && (!strncmp(Doc->Buf.buf+Doc->Buf.size-128, "TAG", 3)))
		get_tag(Doc);
	
	return DPS_OK;
}
