/* Copyright (C) 2003-2004 Datapark corp. All rights reserved.
   Copyright (C) 2000-2002 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "dps_config.h"

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <signal.h>
#include <assert.h>

#ifdef WIN32
#include <time.h>
#endif

#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif

#ifdef HAVE_LIBUTIL_H
#include <libutil.h>
#endif

#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif

#include "dps_common.h"
#include "dps_utils.h"
#include "dps_log.h"
#include "dps_conf.h"
#include "dps_indexer.h"
#include "dps_robots.h"
#include "dps_db.h"
#include "dps_url.h"
#include "dps_parser.h"
#include "dps_proto.h"
#include "dps_hrefs.h"
#include "dps_mutex.h"
#include "dps_crc32.h"
#include "dps_hash.h"
#include "dps_xmalloc.h"
#include "dps_http.h"
#include "dps_host.h"
#include "dps_server.h"
#include "dps_alias.h"
#include "dps_word.h"
#include "dps_crossword.h"
#include "dps_parsehtml.h"
#include "dps_parsexml.h"
#include "dps_spell.h"
#include "dps_execget.h"
#include "dps_agent.h"
#include "dps_match.h"
#include "dps_doc.h"
#include "dps_result.h"
#include "dps_parsedate.h"
#include "dps_unicode.h"
#include "dps_contentencoding.h"
#include "dps_vars.h"
#include "dps_guesser.h"
#include "dps_textlist.h"
#include "dps_id3.h"
#include "dps_stopwords.h"
#include "dps_wild.h"
#include "dps_image.h"
#ifdef HAVE_ZLIB
#include "dps_store.h"
#endif

/* This should be last include */
#ifdef DMALLOC
#include "dmalloc.h"
#endif

#define DPS_THREADINFO(A,s,m)	if(A->Conf->ThreadInfo)A->Conf->ThreadInfo(A,s,m)

/***************************************************************************/

#define MAXHSIZE	1023*4	/* TUNE */

__C_LINK int __DPSCALL DpsStoreHrefs(DPS_AGENT * Indexer) {
	size_t		i, res;
	int             collect_links;
	DPS_DOCUMENT	Doc;

	collect_links = Indexer->Flags.collect_links;
	DpsDocInit(&Doc);

	for (i = 0; i < Indexer->Hrefs.dhrefs; i++) {
		DPS_HREF *H = &Indexer->Hrefs.Href[i];
		if(H->stored == 0) {
			if( 1 /*strlen(H->url) <= DPS_URLSIZE */ ) {  /* FIXME: replace this by config parameter chacking */
				DpsVarListReplaceInt(&Doc.Sections, "Referrer-ID", H->referrer);
				DpsVarListReplaceUnsigned(&Doc.Sections,"Hops", H->hops);
				DpsVarListReplaceStr(&Doc.Sections,"URL",H->url?H->url:"");
				DpsVarListReplaceInt(&Doc.Sections, "URL_ID", DpsStrHash32(H->url ? H->url : ""));
				DpsVarListReplaceInt(&Doc.Sections,"Site_id", H->site_id);
				DpsVarListReplaceInt(&Doc.Sections,"Server_id", H->server_id);
				Doc.charset_id = H->charset_id;
				if (collect_links)
				  if(DPS_OK != (res = DpsURLAction(Indexer, &Doc, DPS_URL_ACTION_ADD_LINK))){
					return(res);
				  }
			}
			H->stored=1;
		}
	}
	for (i = Indexer->Hrefs.dhrefs; i < Indexer->Hrefs.nhrefs; i++) {
		DPS_HREF *H = &Indexer->Hrefs.Href[i];
		if(H->stored == 0) {
			if( 1 /*strlen(H->url) <= DPS_URLSIZE */ ) {  /* FIXME: replace this by config parameter chacking */
				DpsVarListReplaceInt(&Doc.Sections, "Referrer-ID", H->referrer);
				DpsVarListReplaceUnsigned(&Doc.Sections,"Hops", H->hops);
				DpsVarListReplaceStr(&Doc.Sections,"URL",H->url?H->url:"");
				DpsVarListReplaceInt(&Doc.Sections, "URL_ID", DpsStrHash32(H->url ? H->url : ""));
				DpsVarListReplaceInt(&Doc.Sections,"Site_id", H->site_id);
				DpsVarListReplaceInt(&Doc.Sections,"Server_id", H->server_id);
				Doc.charset_id = H->charset_id;
				if(DPS_OK != (res = DpsURLAction(Indexer, &Doc, DPS_URL_ACTION_ADD))){
					return(res);
				}
			}
			H->stored=1;
		}
	}
	DpsDocFree(&Doc);
	
	/* Remember last stored URL num */
	/* Note that it will became 0   */
	/* after next sort in AddUrl    */
	Indexer->Hrefs.dhrefs = Indexer->Hrefs.nhrefs;
	
	/* We should not free URL list with onw database */
	/* to avoid double indexing of the same document */
	/* So, do it if compiled with SQL only           */
	
	/* FIXME: this is incorrect with both SQL and built-in compiled */
	if(Indexer->Hrefs.nhrefs > MAXHSIZE)
		DpsHrefListFree(&Indexer->Hrefs);

	return DPS_OK;
}

static void RelLink(DPS_URL *curURL, DPS_URL *newURL, char **str) {
	const char	*schema = newURL->schema ? newURL->schema : curURL->schema;
	const char	*hostinfo = newURL->hostinfo ? newURL->hostinfo : curURL->hostinfo;
	const char	*path = (newURL->path && newURL->path[0]) ? newURL->path : curURL->path;
	const char	*fname = ((newURL->filename && newURL->filename[0]) || (newURL->path && newURL->path[0])) ? 
	  newURL->filename : curURL->filename;
	const char     *query_string = newURL->query_string;
	char		*pathfile = (char*)DpsMalloc(strlen(DPS_NULL2EMPTY(path)) + strlen(DPS_NULL2EMPTY(fname)) +
						     strlen(DPS_NULL2EMPTY(query_string)) + 5);

	if (newURL->hostinfo == NULL) newURL->charset_id = curURL->charset_id;
	
	if (pathfile == NULL) return;
	sprintf(pathfile, "/%s%s%s",  DPS_NULL2EMPTY(path), DPS_NULL2EMPTY(fname), DPS_NULL2EMPTY(query_string));
	
	DpsURLNormalizePath(pathfile);

	if (!strcasecmp(DPS_NULL2EMPTY(schema), "mailto") || !strcasecmp(DPS_NULL2EMPTY(schema), "javascript")) {
	        *str = (char*)DpsMalloc(strlen(DPS_NULL2EMPTY(schema)) + strlen(DPS_NULL2EMPTY(newURL->specific)) + 4);
		if (*str == NULL) return;
		sprintf(*str, "%s:%s", DPS_NULL2EMPTY(schema), DPS_NULL2EMPTY(newURL->specific));
	} else if(/*!strcasecmp(DPS_NULL2EMPTY(schema), "file") ||*/ !strcasecmp(DPS_NULL2EMPTY(schema), "htdb")) {
	        *str = (char*)DpsMalloc(strlen(DPS_NULL2EMPTY(schema)) + strlen(pathfile) + 4);
		if (*str == NULL) return;
		sprintf(*str, "%s:%s", DPS_NULL2EMPTY(schema), pathfile);
	}else{
	        *str = (char*)DpsMalloc(strlen(DPS_NULL2EMPTY(schema)) + strlen(pathfile) + strlen(DPS_NULL2EMPTY(hostinfo)) + 8);
		if (*str == NULL) return;
		sprintf(*str, "%s://%s%s", DPS_NULL2EMPTY(schema), DPS_NULL2EMPTY(hostinfo), pathfile);
	}
	
	if(!strncmp(*str, "ftp://", 6) && (strstr(*str, ";type=")))
		*(strstr(*str, ";type")) = '\0';
	DPS_FREE(pathfile);
}

static int DpsDocBaseHref(DPS_AGENT *Indexer,DPS_DOCUMENT *Doc){
	const char	*basehref=DpsVarListFindStr(&Doc->Sections,"base.href",NULL);
	/* <BASE HREF="xxx"> stuff            */
	/* Check that URL is properly formed  */
	/* baseURL is just temporary variable */
	/* If parsing  fails we'll use old    */
	/* base href, passed via CurURL       */
	/* Note that we will not check BASE     */
	/* if delete_no_server is unset         */
	/* This is  actually dirty hack. We     */
	/* must check that hostname is the same */
	
	if(basehref /*&& (Doc->Spider.follow==DPS_FOLLOW_WORLD)*/){
		DPS_URL		*baseURL = DpsURLInit(NULL);
		int		parse_res;

		if (baseURL == NULL) return DPS_ERROR;
		if(!(parse_res=DpsURLParse(baseURL, basehref))) {
			DpsURLParse(&Doc->CurURL,basehref);
			DpsLog(Indexer, DPS_LOG_DEBUG, "BASE HREF '%s'", basehref);
		}else{
			switch(parse_res){
			case DPS_URL_LONG:
				DpsLog(Indexer,DPS_LOG_ERROR,"BASE HREF too long: '%s'",basehref);
				break;
			case DPS_URL_BAD:
			default:
				DpsLog(Indexer,DPS_LOG_ERROR,"Error in BASE HREF URL: '%s'",basehref);
			}
		}
		DpsURLFree(baseURL);
	}
	return DPS_OK;
}

#define NS 10
static int DpsFilterFind(DPS_MATCHLIST *L,const char *newhref,char *reason){
	DPS_MATCH_PART	P[NS];
	DPS_MATCH	*M;
	int		res=DPS_METHOD_GET;
	
#ifdef WITH_PARANOIA
	void *paran = DpsViolationEnter();
#endif
	if((M=DpsMatchListFind(L, newhref, NS, P))) {
		sprintf(reason,"%s %s %s '%s'",M->arg,DpsMatchTypeStr(M->match_type),M->case_sense?"Sensitive":"InSensitive",M->pattern);
		res=DpsMethod(M->arg);
	}else{
		sprintf(reason,"Allow by default");
	}
#ifdef WITH_PARANOIA
	DpsViolationExit(paran);
#endif
	return res;
}

int DpsConvertHref(DPS_AGENT *Indexer, DPS_URL *CurURL, DPS_HREF *Href){
	int		parse_res, cascade;
	DPS_URL		*newURL;
	char		*newhref = NULL;
	DPS_MATCH	*Alias;
	char		*alias = NULL;
	size_t		aliassize, nparts = 10;
	DPS_MATCH_PART	Parts[10];
	DPS_SERVER	*Srv;
	DPS_ROBOT_RULE	*rule;
	char		reason[1024]="";

#ifdef WITH_PARANOIA
	void *paran = DpsViolationEnter();
#endif
	if ((newURL = DpsURLInit(NULL)) == NULL) {
	  return DPS_ERROR;
	}
	
	if((parse_res=DpsURLParse(newURL, Href->url))){
		switch(parse_res){
			case DPS_URL_LONG:
				DpsLog(Indexer,DPS_LOG_DEBUG,"URL too long: '%s'",Href->url);
				break;
			case DPS_URL_BAD:
			default:
				DpsLog(Indexer,DPS_LOG_DEBUG,"Error in URL: '%s'",Href->url);
		}
	}

	newURL->charset_id = Href->charset_id;
	RelLink(CurURL, newURL, &newhref);
	
	DpsLog(Indexer,DPS_LOG_DEBUG,"Link '%s' %s",Href->url,newhref);
	for(cascade = 0; ((Alias=DpsMatchListFind(&Indexer->Conf->ReverseAliases,newhref,nparts,Parts))) && (cascade < 1024); cascade++) {
	        aliassize = strlen(Alias->arg) + strlen(Alias->pattern) + strlen(newhref) + 8;
		alias = (char*)DpsRealloc(alias, aliassize);
		if (alias == NULL) {
		  DpsLog(Indexer, DPS_LOG_ERROR, "No memory (%d bytes). %s line %d", aliassize, __FILE__, __LINE__);
		  goto ret;
		}
		DpsMatchApply(alias,aliassize,newhref,Alias->arg,Alias,nparts,Parts);
		if(alias[0]){
		  DpsLog(Indexer,DPS_LOG_DEBUG,"ReverseAlias%d: '%s'", cascade, alias);
		  DPS_FREE(newhref);
		  newhref = (char*)DpsStrdup(alias);
		} else break;
	}
	
		
	DpsURLParse(newURL, newhref);
	
	Href->site_id = 0;

	/* Check Allow/Disallow/CheckOnly stuff */
	Href->method = DpsFilterFind(&Indexer->Conf->Filters, newhref, reason);
	if(Href->method == DPS_METHOD_DISALLOW) {
		DpsLog(Indexer, DPS_LOG_DEBUG, "%s, skip it", reason);
		goto ret;
	}else{
		DpsLog(Indexer, DPS_LOG_DEBUG, "%s", reason);
	}
	
	if (!strcasecmp(DPS_NULL2EMPTY(newURL->schema), "mailto") || !strcasecmp(DPS_NULL2EMPTY(newURL->schema), "javascript")) {
		DpsLog(Indexer,DPS_LOG_DEBUG,"'%s' schema, skip it", newURL->schema, newhref);
		Href->method=DPS_METHOD_DISALLOW;
		goto ret;
	}
	
	if(!(Srv = DpsServerFind(Indexer, &Indexer->Conf->Servers, newhref, newURL->charset_id, NULL))) {
		DpsLog(Indexer,DPS_LOG_DEBUG,"no Server, skip it",newhref);
		Href->method=DPS_METHOD_DISALLOW;
		goto ret;
	}
	DPS_GETLOCK(Indexer, DPS_LOCK_RESOLV);
	rule = DpsRobotRuleFind(Indexer, &Indexer->Conf->Robots, newURL);
	DPS_RELEASELOCK(Indexer, DPS_LOCK_RESOLV);
	if (rule) {
	  DpsLog(Indexer, DPS_LOG_DEBUG, "robots.txt: '%s %s'", (rule->cmd==DPS_METHOD_DISALLOW) ? "Disallow" : "Allow", rule->path);
	  if ((rule->cmd == DPS_METHOD_DISALLOW) || (rule->cmd == DPS_METHOD_VISITLATER) ) {
	    Href->method = rule->cmd;
	    goto ret;
	  }
	}

	if (Href->hops > Srv->MaxHops) {
		DpsLog(Indexer, DPS_LOG_DEBUG, "too many hops (%d), skip it", Href->hops);
		Href->method = DPS_METHOD_DISALLOW;
		goto ret;
	}

	DpsLog(Indexer, DPS_LOG_DEBUG, " Server applied: site_id: %d URL: %s", Srv->site_id, Srv->Match.pattern);
	
	DPS_FREE(Href->url);
	Href->url = (char*)DpsStrdup(newhref);
	Href->server_id = Srv->site_id;
ret:	
	DPS_FREE(newhref);
	DPS_FREE(alias);
	DpsURLFree(newURL);
#ifdef WITH_PARANOIA
	DpsViolationExit(paran);
#endif
	return DPS_OK;
}

static int DpsDocConvertHrefs(DPS_AGENT *Indexer, DPS_DOCUMENT *Doc){
	size_t		i;
	int		hops=DpsVarListFindInt(&Doc->Sections,"Hops",0);
	urlid_t		url_id = (urlid_t)DpsVarListFindInt(&Doc->Sections, "ID", 0);
	uint4           maxhops = DpsVarListFindUnsigned(&Doc->Sections, "MaxHops", 255);

	for(i=0;i<Doc->Hrefs.nhrefs;i++){
		DPS_HREF	*Href=&Doc->Hrefs.Href[i];
		Href->hops=hops+1;
		Href->charset_id = Doc->charset_id;
		DpsConvertHref(Indexer,&Doc->CurURL,Href);
		Href->referrer=url_id;
		if (maxhops > Href->hops) {
		  Href->stored = 0;
		} else {
		  Href->stored = 1;
		  Href->method = DPS_METHOD_DISALLOW;
		}
	}
	return DPS_OK;
}

int DpsDocStoreHrefs(DPS_AGENT *Indexer, DPS_DOCUMENT *Doc){
	size_t i;
	
	if(Doc->method==DPS_METHOD_HEAD)
		return DPS_OK;
	
	DpsDocBaseHref(Indexer,Doc);
	DpsDocConvertHrefs(Indexer,Doc);
	for(i=0;i<Doc->Hrefs.nhrefs;i++){
		DPS_HREF	*Href=&Doc->Hrefs.Href[i];
		if(Href->method!=DPS_METHOD_DISALLOW) {
		  Href->charset_id = Doc->charset_id;
		  DpsHrefListAdd(&Indexer->Hrefs, Href);
		}
	}
	return DPS_OK;
}

/*********************** 'UrlFile' stuff (for -f option) *******************/

__C_LINK int __DPSCALL DpsURLFile(DPS_AGENT *Indexer, const char *fname,int action){
	FILE *url_file;
	char str[1024]="";
	char str1[1024]="";
	int result,res;
	DPS_URL *myurl;
	DPS_HREF Href;

#ifdef WITH_PARANOIA
	void *paran = DpsViolationEnter();
#endif

	if ((myurl = DpsURLInit(NULL)) == NULL) {
	  return DPS_ERROR;
	}
	
	/* Read lines and clear/insert/check URLs                     */
	/* We've already tested in main.c to make sure it can be read */
	/* FIXME !!! Checking should be done here surely              */
	
	if(!strcmp(fname,"-"))
		url_file=stdin;
	else
		url_file=fopen(fname,"r");
	
	while(fgets(str1,sizeof(str1),url_file)){
		char *end;
		if(!str1[0])continue;
		end=str1+strlen(str1)-1;
		while((end>=str1)&&(*end=='\r'||*end=='\n')){
			*end=0;if(end>str1)end--;
		}
		if(!str1[0])continue;
		if(str1[0]=='#')continue;

		if(*end=='\\'){
			*end=0;strcat(str,str1);
			continue;
		}
		strcat(str,str1);
		strcpy(str1,"");

		switch(action){
		case DPS_URL_FILE_REINDEX:
			DpsVarListReplaceStr(&Indexer->Conf->Vars, "ul", str);
			result = DpsURLAction(Indexer, NULL, DPS_URL_ACTION_EXPIRE);
			if(result!=DPS_OK) { DpsURLFree(myurl); 
#ifdef WITH_PARANOIA
			DpsViolationExit(paran);
#endif
			return(result); }
			DpsVarListDel(&Indexer->Conf->Vars, "ul");
			break;
		case DPS_URL_FILE_CLEAR:
			DpsVarListReplaceStr(&Indexer->Conf->Vars, "ul", str);
			result=DpsClearDatabase(Indexer);
			if(result!=DPS_OK) { DpsURLFree(myurl); 
#ifdef WITH_PARANOIA
			DpsViolationExit(paran);
#endif
			return(DPS_ERROR); }
			DpsVarListDel(&Indexer->Conf->Vars, "ul");
			break;
		case DPS_URL_FILE_INSERT:
			DpsHrefInit(&Href);
			Href.url=str;
			Href.method=DPS_METHOD_GET;
			DpsHrefListAdd(&Indexer->Hrefs, &Href);
			break;
		case DPS_URL_FILE_PARSE:
			res=DpsURLParse(myurl, str);
			if((res != DPS_OK) && (myurl->schema == NULL))
				res=DPS_URL_BAD;
			if(res){
				switch(res){
				case DPS_URL_LONG:
					DpsLog(Indexer,DPS_LOG_ERROR,"URL too long: '%s'",str);
					break;
				case DPS_URL_BAD:
				default:
					DpsLog(Indexer,DPS_LOG_ERROR,"Error in URL: '%s'",str);
				}
				DpsURLFree(myurl);
#ifdef WITH_PARANOIA
				DpsViolationExit(paran);
#endif
				return(DPS_ERROR);
			}
			break;
		}
		str[0]=0;
	}
	if(url_file!=stdin)
		fclose(url_file);
	DpsURLFree(myurl);
#ifdef WITH_PARANOIA
	DpsViolationExit(paran);
#endif
	return(DPS_OK);
}




/*******************************************************************/


static int DpsDocAlias(DPS_AGENT *Indexer,DPS_DOCUMENT *Doc){
	DPS_MATCH	*Alias;
	DPS_MATCH_PART	Parts[10];
	size_t		alstrlen, nparts=10;
	const char	*alias_prog=DpsVarListFindStr(&Indexer->Conf->Vars,"AliasProg",NULL);
	char		*alstr;
	int		result=DPS_OK;
	const char	*url=DpsVarListFindStr(&Doc->Sections,"URL","");

	alstrlen = 256 + strlen(url);
	if ((alstr = (char*)DpsMalloc(alstrlen)) == NULL) return DPS_ERROR;
	alstr[0] = '\0';
	if(alias_prog){
		result = DpsAliasProg(Indexer, alias_prog, url, alstr, alstrlen - 1);
		DpsLog(Indexer,DPS_LOG_EXTRA,"AliasProg result: '%s'",alstr);
		if(result!=DPS_OK || alstr[0]) { DPS_FREE(alstr); return result; }
	}
	
	/* Find alias when aliastr is empty, i.e.     */
	/* when there is no alias in "Server" command */
	/* and no AliasProg                           */
	if((Alias=DpsMatchListFind(&Indexer->Conf->Aliases,url,nparts,Parts))){
		DpsMatchApply(alstr, alstrlen - 1, url, Alias->arg, Alias, nparts, Parts);
	}
	if(alstr[0]){
		DpsVarListReplaceStr(&Doc->Sections,"Alias",alstr);
	}
	DPS_FREE(alstr);
	return DPS_OK;
}




static int DpsDocCheck(DPS_AGENT *Indexer, DPS_SERVER *CurSrv, DPS_DOCUMENT *Doc) {
	char		reason[1024]="";
	int		nerrors=-1;
	int		hops=DpsVarListFindInt(&Doc->Sections,"Hops",0);
	const char	*method=DpsVarListFindStr(&CurSrv->Vars,"Method","Allow");
	const int       older = DpsVarListFindInt(&Doc->Sections, "DeleteOlder", 0);
	
	switch(CurSrv->Match.match_type){
		case DPS_MATCH_WILD:
			DpsLog(Indexer,DPS_LOG_DEBUG, "Realm %s wild '%s'", method, CurSrv->Match.pattern);
			break;
		case DPS_MATCH_REGEX:
			DpsLog(Indexer,DPS_LOG_DEBUG, "Realm %s regex '%s'", method, CurSrv->Match.pattern);
			break;
		case DPS_MATCH_SUBNET:
			DpsLog(Indexer,DPS_LOG_DEBUG, "Subnet %s '%s'", method, CurSrv->Match.pattern);
			break;
		case DPS_MATCH_BEGIN:
		default:
			DpsLog(Indexer,DPS_LOG_DEBUG, "Server %s '%s'", method, CurSrv->Match.pattern);
			break;
	}
	
	if((Doc->method=DpsMethod(method)) == DPS_METHOD_GET){
		/* Check Allow/Disallow/CheckOnly stuff */
		Doc->method=DpsFilterFind(&Indexer->Conf->Filters,DpsVarListFindStr(&Doc->Sections,"URL",""),reason);
		DpsLog(Indexer,DPS_LOG_DEBUG,"%s",reason);
	}
	
	if(Doc->method==DPS_METHOD_DISALLOW)return DPS_OK;
	
	/* Check that hops is less than MaxHops */
	if(hops>Doc->Spider.maxhops){
		DpsLog(Indexer,DPS_LOG_WARN,"Too many hops (%d)",hops);
		Doc->method=DPS_METHOD_DISALLOW;
		return DPS_OK;
	}

	/* Check for older focs */
	if (older > 0) {
	  time_t now = time(NULL), last_mod_time = DpsHttpDate2Time_t(DpsVarListFindStr(&Doc->Sections, "Last-Modified", ""));

	  if (last_mod_time > 0) {
	    if ((int)(now - last_mod_time) > older) {
	      DpsLog(Indexer, DPS_LOG_WARN, "Too old document (%d > %d)", now - last_mod_time, older);
	      Doc->method = DPS_METHOD_DISALLOW;
	      return DPS_OK;
	    }
	  } else {
	    time_t since = (time_t)DpsVarListFindInt(&Doc->Sections, "Since", 0);
	    if ((int)(now - since) > older) {
	      DpsLog(Indexer, DPS_LOG_WARN, "Too old document (%d > %d)", now - since, older);
	      Doc->method = DPS_METHOD_DISALLOW;
	      return DPS_OK;
	    }
	  }
	}
	
	/* Check for too many errors on this server */
	DpsDocLookupConn(Indexer, Doc);
	nerrors = (Doc->connp.Host != NULL) ?  Doc->connp.Host->net_errors : 0;
	
	if((nerrors>=Doc->Spider.max_net_errors)&&(Doc->Spider.max_net_errors)){
		size_t	next_index_time=time(NULL)+Doc->Spider.net_error_delay_time;
		char	buf[64];
		
		DpsLog(Indexer,DPS_LOG_WARN,"Too many network errors (%d) for this server",nerrors);
		DpsVarListReplaceInt(&Doc->Sections,"Status",DPS_HTTP_STATUS_SERVICE_UNAVAILABLE);
		DpsTime_t2HttpStr((int)next_index_time, buf);
		DpsVarListReplaceStr(&Doc->Sections,"Next-Index-Time",buf);
		Doc->method=DPS_METHOD_VISITLATER;
		return DPS_OK;
	}
	
	return DPS_OK;
}





static int DpsDocParseContent(DPS_AGENT * Indexer, DPS_DOCUMENT * Doc){
	
#ifdef WITH_PARSER
	DPS_PARSER	*Parser;
	const int       status = DpsVarListFindInt(&Doc->Sections, "Status", 0);
#endif
	const char	*real_content_type=NULL;
	const char	*url=DpsVarListFindStr(&Doc->Sections,"URL","");
	const char	*ct=DpsVarListFindStr(&Doc->Sections,"Content-Type","");
	const char	*ce=DpsVarListFindStr(&Doc->Sections,"Content-Encoding","");
	int		result=DPS_OK;
	
	if(!strcmp(DPS_NULL2EMPTY(Doc->CurURL.filename), "robots.txt")) return DPS_OK;
	
#ifdef HAVE_ZLIB
	if(!strcasecmp(ce,"gzip") || !strcasecmp(ce,"x-gzip")){
		DPS_THREADINFO(Indexer,"UnGzip",url);
		DpsUnGzip(Doc);
		DpsVarListReplaceInt(&Doc->Sections, "Content-Length", Doc->Buf.buf - Doc->Buf.content + (int)Doc->Buf.size);
	}else
	if(!strcasecmp(ce,"deflate")){
		DPS_THREADINFO(Indexer,"Inflate",url);
		DpsInflate(Doc);
		DpsVarListReplaceInt(&Doc->Sections, "Content-Length", Doc->Buf.buf - Doc->Buf.content + (int)Doc->Buf.size);
	}else
	if(!strcasecmp(ce,"compress") || !strcasecmp(ce,"x-compress")){
		DPS_THREADINFO(Indexer,"Uncompress",url);
		DpsUncompress(Doc);
		DpsVarListReplaceInt(&Doc->Sections, "Content-Length", Doc->Buf.buf - Doc->Buf.content + (int)Doc->Buf.size);
	}else
#endif	
	if(!strcasecmp(ce,"identity") || !strcasecmp(ce,"")){
		/* Nothing to do*/
	}else{
		DpsLog(Indexer,DPS_LOG_ERROR,"Unsupported Content-Encoding");
		DpsVarListReplaceInt(&Doc->Sections,"Status",DPS_HTTP_STATUS_UNSUPPORTED_MEDIA_TYPE);
	}
	
#ifdef WITH_PARSER
	/* Let's try to start external parser for this Content-Type */

	if((Parser = DpsParserFind(&Indexer->Conf->Parsers, ct))) {
		DpsLog(Indexer,DPS_LOG_DEBUG,"Found external parser '%s' -> '%s'",
			Parser->from_mime?Parser->from_mime:"NULL",
			Parser->to_mime?Parser->to_mime:"NULL");
	}
	if(Parser) {
	  if (status == DPS_HTTP_STATUS_OK) {
	    if (DpsParserExec(Indexer, Parser, Doc)) {
		char *to_charset;
		real_content_type=Parser->to_mime?Parser->to_mime:"unknown";
		DpsLog(Indexer,DPS_LOG_DEBUG,"Parser-Content-Type: %s",real_content_type);
		if((to_charset=strstr(real_content_type,"charset="))){
		        const char *cs = DpsCharsetCanonicalName(DpsTrim(to_charset + 8, " \t"));
			DpsVarListReplaceStr(&Doc->Sections, "Server-Charset", cs);
			DpsLog(Indexer,DPS_LOG_DEBUG, "to_charset='%s'", cs);
		}
#ifdef DEBUG_PARSER
		fprintf(stderr,"content='%s'\n",Doc->content);
#endif
	    }
	  } else {
	    DpsLog(Indexer, DPS_LOG_WARN, "Parser not executed, document status: %d", status);
	    return result;
	  }
	}
#endif
	
	if(!real_content_type)real_content_type=ct;
	DpsVarListAddStr(&Doc->Sections,"Parser-Content-Type",real_content_type);
	
#ifdef HAVE_ZLIB
	if ( strncmp(real_content_type, "text/", 5) == 0) {
	  DpsStoreDoc(Indexer, Doc);
	}
	
#endif
	
	/* CRC32 without headers */
	{
	  size_t crclen=Doc->Buf.size - (Doc->Buf.content-Doc->Buf.buf);
	  DpsVarListReplaceInt(&Doc->Sections, "crc32", (int)DpsCRC32(Doc->Buf.content,crclen));
	}

	if(Doc->method!=DPS_METHOD_HEAD){
		if(!strncasecmp(real_content_type,"text/plain",10)){
			DpsParseText(Indexer,Doc);
		}else
		if(!strncasecmp(real_content_type,"text/tab-separated-values",25)){
			DpsParseText(Indexer,Doc);
		}else
		if(!strncasecmp(real_content_type,"text/css",8)){
			DpsParseText(Indexer,Doc);
		}else	
		if(!strncasecmp(real_content_type,"text/html",9)){
			DpsHTMLParse(Indexer,Doc);
		}else
#ifdef WITH_XML
		if(!strncasecmp(real_content_type, "text/xml",8)) {
			DpsXMLParse(Indexer, Doc);
		}else
#endif
		if(!strncasecmp(real_content_type, "audio/mpeg",10) || !strncasecmp(real_content_type, "audio/x-mpeg",12)) {
			DpsMP3Parse(Indexer,Doc);
		}else
		if(!strncasecmp(real_content_type, "image/gif", 9)) {
			DpsGIFParse(Indexer, Doc);
		}else
{
			/* Unknown Content-Type  */
			DpsLog(Indexer,DPS_LOG_ERROR,"Unsupported Content-Type '%s'",real_content_type);
			DpsVarListReplaceInt(&Doc->Sections,"Status",DPS_HTTP_STATUS_UNSUPPORTED_MEDIA_TYPE);
		}
	}
	return result;
}



static int DpsNextTarget(DPS_AGENT * Indexer,DPS_DOCUMENT *Result){
	int	result=DPS_NOTARGET;
	int     u;
	
	DPS_GETLOCK(Indexer, DPS_LOCK_CONF);
	u = (Indexer->Conf->url_number <= 0);
	DPS_RELEASELOCK(Indexer, DPS_LOCK_CONF);

	if (u) {
		return DPS_NOTARGET;
	}
	
	/* Load targets into memory cache */
	DPS_GETLOCK(Indexer, DPS_LOCK_THREAD);
	u = (Indexer->Conf->Targets.cur_row >= Indexer->Conf->Targets.num_rows);
	if (u) {
		result = DpsTargets(Indexer);
		if(result!=DPS_OK) {
		  return result;
		}
	}
	
	/* Choose next target */
/*	DPS_GETLOCK(Indexer, DPS_LOCK_CONF);*/
	if(Indexer->Conf->Targets.num_rows && (Indexer->Conf->Targets.cur_row < Indexer->Conf->Targets.num_rows ) ) {
		DPS_DOCUMENT *Doc=&Indexer->Conf->Targets.Doc[Indexer->Conf->Targets.cur_row];

		if (Doc != NULL) {
		  DpsVarListReplaceLst(&Result->Sections, &Indexer->Conf->Sections, NULL, "*");
		  DpsVarListReplaceLst(&Result->Sections, &Doc->Sections, NULL, "*");
		  DpsVarListReplaceLst(&Result->RequestHeaders, &Doc->RequestHeaders, NULL, "*");
		  Result->charset_id = Doc->charset_id;
		  
		  Indexer->Conf->Targets.cur_row++;
		  Indexer->Conf->url_number--;
		  DPS_RELEASELOCK(Indexer, DPS_LOCK_THREAD);
/*		  DPS_RELEASELOCK(Indexer, DPS_LOCK_CONF);*/
		  DpsVarListLog(Indexer, &Result->Sections, DPS_LOG_DEBUG, "Target");
		  return DPS_OK;
		}
	}
	
	DPS_RELEASELOCK(Indexer, DPS_LOCK_THREAD);
/*	DPS_RELEASELOCK(Indexer, DPS_LOCK_CONF);*/
	return DPS_NOTARGET;
}

static int DpsVarList2Spider(DPS_SPIDERPARAM *S,DPS_VARLIST *V){
	S->period		=DpsVarListFindInt(V, "Period",		DPS_DEFAULT_REINDEX_TIME);
	S->maxhops		=DpsVarListFindInt(V, "MaxHops",	DPS_DEFAULT_MAX_HOPS);
	S->follow		=DpsVarListFindInt(V, "Follow",		DPS_FOLLOW_PATH);
	S->max_net_errors	=DpsVarListFindInt(V, "MaxNetErrors",	DPS_MAXNETERRORS);
	S->net_error_delay_time	=DpsVarListFindInt(V, "NetErrorDelayTime",DPS_DEFAULT_NET_ERROR_DELAY_TIME);
	S->read_timeout		=DpsVarListFindInt(V, "ReadTimeOut",	DPS_READ_TIMEOUT);
	S->doc_timeout		=DpsVarListFindInt(V, "DocTimeOut",	DPS_DOC_TIMEOUT);
	S->index		=DpsVarListFindInt(V, "Index",		1);
	S->use_robots		=DpsVarListFindInt(V, "Robots",		1);
	S->use_clones		=DpsVarListFindInt(V, "DetectClones",	1);
	return DPS_OK;
}

__C_LINK int __DPSCALL DpsIndexNextURL(DPS_AGENT *Indexer){
	int		result=DPS_OK;
	DPS_DOCUMENT	Doc;
	const char	*url, *alias;
	char		*origurl = NULL, *aliasurl = NULL;
	DPS_SERVER	*Server = NULL;
#ifdef WITH_PARANOIA
	void * paran = DpsViolationEnter();
#endif

	TRACE_IN(Indexer, "DpsIndexNextURL");
	
	DpsDocInit(&Doc);
	
	DPS_THREADINFO(Indexer,"Selecting","");

	if(DPS_OK==(result=DpsStoreHrefs(Indexer))) {
	        if (Indexer->action != DPS_OK) {
		  TRACE_OUT(Indexer);
#ifdef WITH_PARANOIA
		  DpsViolationExit(paran);
#endif
		  return Indexer->action;
		}
		result = DpsNextTarget(Indexer, &Doc);
	}
	
	if(result==DPS_NOTARGET){
		DpsDocFree(&Doc);	/* To free Doc.connp->connp */
		TRACE_OUT(Indexer);
#ifdef WITH_PARANOIA
		DpsViolationExit(paran);
#endif
		return result;
	}
	
	if(result!=DPS_OK){
		DpsDocFree(&Doc);
		TRACE_OUT(Indexer);
#ifdef WITH_PARANOIA
		DpsViolationExit(paran);
#endif
		return result;
	}
	
	url=DpsVarListFindStr(&Doc.Sections,"URL","");
	DpsVarListReplaceInt(&Doc.Sections,"crc32old",DpsVarListFindInt(&Doc.Sections,"crc32",0));
	DpsLog(Indexer,DPS_LOG_INFO,"URL: %s",url);
	
#ifdef HAVE_SETPROCTITLE
	/* To see the URL being indexed in "ps" output on xBSD */
	setproctitle("[%d] URL:%s", Indexer->handle, url);
#endif
	
	
	/* Collect information from Conf */
/*	DPS_GETLOCK(Indexer,DPS_LOCK_CONF);*/
	
	if(!Doc.Buf.buf){
		/* Alloc buffer for document */
	        DPS_GETLOCK(Indexer,DPS_LOCK_CONF);
		Doc.Buf.maxsize=(size_t)DpsVarListFindInt(&Indexer->Conf->Vars,"MaxDocSize",DPS_MAXDOCSIZE);
		DPS_RELEASELOCK(Indexer,DPS_LOCK_CONF);
		if ((Doc.Buf.buf=(char*)DpsMalloc(Doc.Buf.maxsize + 1)) == NULL) {
		  DpsLog(Indexer, DPS_LOG_ERROR, "Out of memory (%d bytes) %s:%d", Doc.Buf.maxsize + 1, __FILE__, __LINE__);
		  TRACE_OUT(Indexer);
		  return DPS_ERROR;
		}
		Doc.Buf.buf[0]='\0';
	}
	
	/* Check that URL has valid syntax */
	if(DpsURLParse(&Doc.CurURL,url)){
		DpsLog(Indexer,DPS_LOG_WARN,"Invalid URL: %s",url);
		Doc.method = DPS_METHOD_DISALLOW;
	}else
	if ((Doc.CurURL.filename != NULL) && (!strcmp(Doc.CurURL.filename, "robots.txt"))) {
		Doc.method = DPS_METHOD_DISALLOW;
	}else{
		char		*alstr = NULL;

		Doc.CurURL.charset_id = Doc.charset_id;
		/* Find correspondent Server */
/*		DPS_GETLOCK(Indexer,DPS_LOCK_CONF);*/
		Server = DpsServerFind(Indexer, &Indexer->Conf->Servers, url, Doc.charset_id, &alstr);
/*		DPS_RELEASELOCK(Indexer,DPS_LOCK_CONF);*/
		if ( !Server ) {
			DpsLog(Indexer,DPS_LOG_WARN,"No 'Server' command for url");
			Doc.method = DPS_METHOD_DISALLOW;
		}else{
		        DPS_GETLOCK(Indexer,DPS_LOCK_CONF);
			Doc.lcs=Indexer->Conf->lcs;
			DpsVarList2Spider(&Doc.Spider,&Server->Vars);
			
			DpsDocAddConfExtraHeaders(Indexer->Conf,&Doc);
			DpsDocAddServExtraHeaders(Server,&Doc);
			DPS_RELEASELOCK(Indexer,DPS_LOCK_CONF);

			DpsVarListReplaceLst(&Doc.Sections, &Server->Vars,NULL,"*");
			DpsVarListReplaceInt(&Doc.Sections, "Site_id", DpsServerGetSiteId(Indexer, Server, &Doc.CurURL));
			DpsVarListReplaceInt(&Doc.Sections, "Server_id", Server->site_id);
			DpsVarListReplaceInt(&Doc.Sections, "MaxHops", Doc.Spider.maxhops);
			
			if(alstr != NULL) {
				/* Server Primary alias found */
				DpsVarListReplaceStr(&Doc.Sections,"Alias",alstr);
			}else{
				/* Apply non-primary alias */
				result=DpsDocAlias(Indexer,&Doc);
			}

			if((alias = DpsVarListFindStr(&Doc.Sections, "Alias", NULL))) {
			  const char *u = DpsVarListFindStr(&Doc.Sections, "URL", NULL);
			  origurl = (char*)DpsStrdup(u);
			  aliasurl = (char*)DpsStrdup(alias);
			  DpsLog(Indexer,DPS_LOG_EXTRA,"Alias: '%s'", alias);
			  DpsVarListReplaceStr(&Doc.Sections, "URL", alias);
			  DpsVarListReplaceInt(&Doc.Sections, "URL_ID", DpsStrHash32(alias));
			  DpsURLParse(&Doc.CurURL, alias);
			}

			/* Check hops, network errors, filters */
			result=DpsDocCheck(Indexer,Server,&Doc);
			
		}
		DPS_FREE(alstr);
	}
/*	DPS_RELEASELOCK(Indexer,DPS_LOCK_CONF);*/
	
	
	if(result!=DPS_OK){
		DpsDocFree(&Doc);
		TRACE_OUT(Indexer);
#ifdef WITH_PARANOIA
		DpsViolationExit(paran);
#endif
		return result;
	}
	
	if(Doc.method!=DPS_METHOD_DISALLOW && Doc.method!=DPS_METHOD_VISITLATER){
		DpsDocAddDocExtraHeaders(Indexer, &Doc);
		if(!strncmp(DPS_NULL2EMPTY(Doc.CurURL.schema), "http", 4)) {
			if(!Doc.Spider.use_robots){
				DpsLog(Indexer,DPS_LOG_WARN, "robots.txt support is disallowed for '%s'", 
				       DPS_NULL2EMPTY(Doc.CurURL.hostinfo));
				DPS_GETLOCK(Indexer,DPS_LOCK_CONF);
				result = DpsRobotParse(Indexer, NULL, NULL, (char*)DPS_NULL2EMPTY(Doc.CurURL.hostinfo));
				DPS_RELEASELOCK(Indexer,DPS_LOCK_CONF);
			}else{
				DPS_ROBOT_RULE	*rule;

				/* Check whether URL is disallowed by robots.txt */
				DPS_GETLOCK(Indexer,DPS_LOCK_RESOLV);
				rule = DpsRobotRuleFind(Indexer, &Indexer->Conf->Robots, &Doc.CurURL);
				DPS_RELEASELOCK(Indexer,DPS_LOCK_RESOLV);
				if(rule) {
					DpsLog(Indexer,DPS_LOG_WARN,"robots.txt: '%s %s'",(rule->cmd==DPS_METHOD_DISALLOW)?"Disallow":"Allow",rule->path);
					if((rule->cmd == DPS_METHOD_DISALLOW) || (rule->cmd == DPS_METHOD_VISITLATER) )
						Doc.method=rule->cmd;
				}
			}
		}
		if(origurl != NULL){
			DpsVarListReplaceStr(&Doc.Sections,"URL",origurl);
			DpsVarListReplaceInt(&Doc.Sections, "URL_ID", DpsStrHash32(origurl));
			DpsURLParse(&Doc.CurURL,origurl);
		}
	}
	
	
	if(result!=DPS_OK){
	        DPS_FREE(origurl); DPS_FREE(aliasurl);
		DpsDocFree(&Doc);
		TRACE_OUT(Indexer);
#ifdef WITH_PARANOIA
		DpsViolationExit(paran);
#endif
		return result;
	}

	if(Doc.method!=DPS_METHOD_DISALLOW && Doc.method!=DPS_METHOD_VISITLATER){
		int	start,state,status=0;
		int	mp3type=DPS_MP3_UNKNOWN;
		
		if(!(Indexer->flags&DPS_FLAG_REINDEX)){
			const char *l=DpsVarListFindStr(&Doc.Sections,"Last-Modified",NULL);
			if (l) DpsVarListReplaceStr(&Doc.RequestHeaders, "If-Modified-Since", l);
		}
		
		DPS_THREADINFO(Indexer,"Getting",url);
		
		start=(Doc.method==DPS_METHOD_CHECKMP3 || Doc.method==DPS_METHOD_CHECKMP3ONLY)?1:0;
		
		for(state=start;state>=0;state--){
			const char	*hdr=NULL;
			
			if(state==1)hdr="bytes=0-256";
			if(mp3type==DPS_MP3_TAG)hdr="bytes=-128";
			
			DpsVarListReplaceInt(&Doc.Sections, "Status", DPS_HTTP_STATUS_UNKNOWN);
			
			if(aliasurl != NULL) {
			  DpsVarListReplaceStr(&Doc.Sections,"URL",alias);
			  DpsVarListReplaceInt(&Doc.Sections, "URL_ID", DpsStrHash32(alias));
			  DpsURLParse(&Doc.CurURL,alias);
			}
				
			DpsVarListLog(Indexer,&Doc.RequestHeaders,DPS_LOG_DEBUG,"Request");
				
			if(hdr) {
			        DpsVarListAddStr(&Doc.RequestHeaders, "Range", hdr);
				DpsLog(Indexer, DPS_LOG_INFO, "Range: [%s]", hdr);
			}
				
			result = DpsGetURL(Indexer, &Doc);
				
			if(hdr) {
			        DpsVarListDel(&Doc.RequestHeaders, "Range");
			}
				
			if(origurl != NULL) {
			  DpsVarListReplaceStr(&Doc.Sections,"URL",origurl);
			  DpsVarListReplaceInt(&Doc.Sections, "URL_ID", DpsStrHash32(origurl));
			  DpsURLParse(&Doc.CurURL,origurl);
			}
			
			if(result!=DPS_OK){
			        DPS_FREE(origurl); DPS_FREE(aliasurl);
				DpsDocFree(&Doc);
				TRACE_OUT(Indexer);
#ifdef WITH_PARANOIA
				DpsViolationExit(paran);
#endif
				return result;
			}
			
			DpsParseHTTPResponse(Indexer,&Doc);
			DpsDocProcessResponseHeaders(Indexer,&Doc);
			DpsVarListLog(Indexer,&Doc.Sections,DPS_LOG_DEBUG,"Response");
			
			status=DpsVarListFindInt(&Doc.Sections,"Status",0);
			
			DpsLog(Indexer, DPS_LOG_EXTRA, "Status: %d %s", status, DpsHTTPErrMsg(status));

			if(status != DPS_HTTP_STATUS_PARTIAL_OK && status != DPS_HTTP_STATUS_OK)
				break;
			
			if(state==1){	/* Needs guessing */
				if(DPS_MP3_UNKNOWN != (mp3type=DpsMP3Type(&Doc))){
					DpsVarListReplaceStr(&Doc.Sections,"Content-Type","audio/mpeg");
					if(Doc.method == DPS_METHOD_CHECKMP3ONLY && mp3type != DPS_MP3_TAG) break;
				}
				if(Doc.method == DPS_METHOD_CHECKMP3ONLY) break;
			}
		}
		
		/* Add URL from Location: header */
		/* This is to give a chance for  */
		/* a concurent thread to take it */
		result=DpsDocStoreHrefs(Indexer,&Doc);
		if(result!=DPS_OK){
		        DPS_FREE(origurl); DPS_FREE(aliasurl);
			DpsDocFree(&Doc);
			TRACE_OUT(Indexer);
#ifdef WITH_PARANOIA
			DpsViolationExit(paran);
#endif
			return result;
		}
		
		/* Increment indexer's download statistics */
		Indexer->nbytes+=Doc.Buf.size;
		Indexer->ndocs++;
		
		if((!Doc.Buf.content) && (status < 500)) {
			DpsLog(Indexer, DPS_LOG_ERROR, "No data received");
			status=DPS_HTTP_STATUS_SERVICE_UNAVAILABLE;
			DpsVarListReplaceInt(&Doc.Sections,"Status",status);
		}
		
		if(status==DPS_HTTP_STATUS_OK || status==DPS_HTTP_STATUS_PARTIAL_OK){
		   	size_t		wordnum, min_size;
			size_t	hdr_len = Doc.Buf.content - Doc.Buf.buf;
			size_t	cont_len = Doc.Buf.size - hdr_len;
			const char *lang = NULL;
			int collect_links, skip_too_small;
		   	
			DPS_GETLOCK(Indexer, DPS_LOCK_CONF);
			min_size = (size_t)DpsVarListFindUnsigned(&Indexer->Conf->Vars, "MinDocSize", 0);
			DPS_RELEASELOCK(Indexer, DPS_LOCK_CONF);
			skip_too_small = (cont_len < min_size);

			if (skip_too_small) {
			  Doc.method = DPS_METHOD_HEAD;
			  DpsLog(Indexer, DPS_LOG_EXTRA, "Too small content size (%d < %d), CheckOnly.", cont_len, min_size); 
			}

			DPS_THREADINFO(Indexer,"Parsing",url);
			
			result = DpsDocParseContent(Indexer, &Doc);
			if(result!=DPS_OK){
			        DPS_FREE(origurl); DPS_FREE(aliasurl);
				DpsDocFree(&Doc);
				TRACE_OUT(Indexer);
#ifdef WITH_PARANOIA
				DpsViolationExit(paran);
#endif
				return result;
			}
			
			DPS_GETLOCK(Indexer,DPS_LOCK_CONF);
			if(Indexer->Conf->LangMaps.nmaps && Doc.method != DPS_METHOD_DISALLOW){
				register size_t t;
				int flag = Indexer->Flags.update_lm;

				bzero((void*)Indexer->LangMap, sizeof(DPS_LANGMAP));
				for (t = 0; t <= DPS_LM_HASHMASK; t++) Indexer->LangMap->memb[t].index = t;
				for(t=0;t<Doc.TextList.nitems;t++){
					DPS_TEXTITEM *Item = &Doc.TextList.Items[t];
					DpsBuildLangMap(Indexer->LangMap, Item->str, strlen(Item->str), flag);
				}
			}

			DPS_RELEASELOCK(Indexer,DPS_LOCK_CONF);
			
			DpsGuessCharSet(Indexer, &Doc, &Indexer->Conf->LangMaps, Indexer->LangMap);
				
			DpsLog(Indexer, DPS_LOG_EXTRA, "Guesser: Lang: %s, Charset: %s",
			       lang = DpsVarListFindStr(&Doc.Sections,"Content-Language",""),
			       DpsVarListFindStr(&Doc.Sections,"Charset",""));
			
			DpsParseURLText(Indexer,&Doc);
/*			DpsParseHeaders(Indexer,&Doc);*/
			if (Doc.method!=DPS_METHOD_HREFONLY) {
				DpsPrepareWords(Indexer,&Doc);
			}
			
			/* Remove StopWords */
			DPS_GETLOCK(Indexer,DPS_LOCK_CONF);
			for(wordnum=0;wordnum<Doc.Words.nwords;wordnum++){
				const int	*w = Doc.Words.Word[wordnum].uword;
				size_t		wlen = Doc.Words.Word[wordnum].ulen;
				
				if(wlen > Indexer->Conf->WordParam.max_word_len ||
				   wlen < Indexer->Conf->WordParam.min_word_len ||
				   DpsStopListFind(&Indexer->Conf->StopWords, w, lang ) != NULL)
				{
					Doc.Words.Word[wordnum].coord=0;
				}	
			}
			for(wordnum=0;wordnum<Doc.CrossWords.ncrosswords;wordnum++){
				const int	*w = Doc.CrossWords.CrossWord[wordnum].uword;
				size_t		wlen = Doc.CrossWords.CrossWord[wordnum].ulen;
				
				if(wlen>Indexer->Conf->WordParam.max_word_len ||
				   wlen<Indexer->Conf->WordParam.min_word_len ||
				   DpsStopListFind(&Indexer->Conf->StopWords,w, lang) != NULL)
				{
					Doc.CrossWords.CrossWord[wordnum].weight=0;
				}	
			}
			collect_links = Indexer->Flags.collect_links;
			DPS_RELEASELOCK(Indexer,DPS_LOCK_CONF);
			if (collect_links)
			  if(DPS_OK != (result = DpsURLAction(Indexer, &Doc, DPS_URL_ACTION_LINKS_DELETE))) {
			    TRACE_OUT(Indexer);
#ifdef WITH_PARANOIA
			    DpsViolationExit(paran);
#endif
			    return result;
			  }
		}
	}
	
	if(DPS_OK==(result=DpsDocStoreHrefs(Indexer,&Doc))){
		if(DPS_OK!=(result=DpsStoreHrefs(Indexer)))
			result=DPS_ERROR;
	}
	
	if(result!=DPS_OK){
	        DPS_FREE(origurl); DPS_FREE(aliasurl);
		DpsDocFree(&Doc);
		TRACE_OUT(Indexer);
#ifdef WITH_PARANOIA
		DpsViolationExit(paran);
#endif
		return result;
	}
	
	/* Free unnecessary information */
	DpsHrefListFree(&Doc.Hrefs);
	DpsVarListFree(&Doc.RequestHeaders);
	DpsTextListFree(&Doc.TextList);
	DPS_FREE(Doc.Buf.buf);
	Doc.Buf.maxsize=0;
	
	result = DpsURLAction(Indexer, &Doc, DPS_URL_ACTION_FLUSH);
	
	DPS_FREE(origurl); DPS_FREE(aliasurl);
	DPS_GETLOCK(Indexer, DPS_LOCK_THREAD);
	if (result == DPS_OK) result = Indexer->action;
	DPS_RELEASELOCK(Indexer, DPS_LOCK_THREAD);
	TRACE_OUT(Indexer);
#ifdef WITH_PARANOIA
	DpsViolationExit(paran);
#endif
	return result;
}
