/* Copyright (C) 2003-2004 Datapark corp. All rights reserved.
   Copyright (C) 2000-2002 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "dps_config.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>

#include "dps_common.h"
#include "dps_robots.h"
#include "dps_utils.h"
#include "dps_vars.h"
#include "dps_log.h"
#include "dps_doc.h"
#include "dps_server.h"
#include "dps_hash.h"
#include "dps_proto.h"
#include "dps_url.h"
#include "dps_mutex.h"
#include "dps_http.h"


DPS_ROBOT* DpsRobotFind(DPS_ROBOTS *Robots,const char *hostinfo){
	size_t i;
	DPS_ROBOT *r;
#ifdef WITH_PARANOIA
	void *paran = DpsViolationEnter();
#endif

	/* FIXMI: <hash> or <btree> or <bsearch with partial sorting> or <?> */
	for(i=0;i<Robots->nrobots;i++){
		if(!strcasecmp(hostinfo, DPS_NULL2EMPTY(Robots->Robot[i].hostinfo))) {
		  r = &Robots->Robot[i];
#ifdef WITH_PARANOIA
		  DpsViolationExit(paran);
#endif
		  return r;
		}
	}
#ifdef WITH_PARANOIA
	DpsViolationExit(paran);
#endif
	return NULL;
}

static DPS_ROBOT* DeleteRobotRules(DPS_ROBOTS *Robots,char *hostinfo){
	DPS_ROBOT *robot;
	size_t i;
#ifdef WITH_PARANOIA
	void *paran = DpsViolationEnter();
#endif
	
	if((robot = DpsRobotFind(Robots, DPS_NULL2EMPTY(hostinfo))) != NULL) {
		for(i=0;i<robot->nrules;i++){
			DPS_FREE(robot->Rule[i].path);
		}
		robot->nrules=0;
		DPS_FREE(robot->Rule);
#ifdef WITH_PARANOIA
		DpsViolationExit(paran);
#endif
		return robot;
	}
#ifdef WITH_PARANOIA
	DpsViolationExit(paran);
#endif
	return NULL;
}

DPS_ROBOT* DpsRobotAddEmpty(DPS_ROBOTS *Robots,char *hostinfo){
  DPS_ROBOT *r;
#ifdef WITH_PARANOIA
	void *paran = DpsViolationEnter();
#endif

	Robots->Robot = (DPS_ROBOT*)DpsRealloc(Robots->Robot, (Robots->nrobots + 1) * sizeof(DPS_ROBOT));
	if(Robots->Robot==NULL) {
	  Robots->nrobots = 0;
#ifdef WITH_PARANOIA
	  DpsViolationExit(paran);
#endif
	  return NULL;
	}

	bzero((void*)&Robots->Robot[Robots->nrobots], sizeof(DPS_ROBOT));
	Robots->Robot[Robots->nrobots].hostinfo = (char*)DpsStrdup(DPS_NULL2EMPTY(hostinfo));
	
	Robots->nrobots++;
	r = &Robots->Robot[Robots->nrobots - 1];
#ifdef WITH_PARANOIA
	DpsViolationExit(paran);
#endif
	return r;
}

static int AddRobotRule(DPS_ROBOT *robot,int cmd,char *path){
#ifdef WITH_PARANOIA
	void *paran = DpsViolationEnter();
#endif
	robot->Rule = (DPS_ROBOT_RULE*)DpsRealloc(robot->Rule, (robot->nrules + 1) * sizeof(DPS_ROBOT_RULE));
	if(robot->Rule==NULL) {
	  robot->nrules = 0;
#ifdef WITH_PARANOIA
	  DpsViolationExit(paran);
#endif
	  return DPS_ERROR;
	}
	
	robot->Rule[robot->nrules].cmd = cmd;
	robot->Rule[robot->nrules].path = (char*)DpsStrdup(DPS_NULL2EMPTY(path));
	
	robot->nrules++;
#ifdef WITH_PARANOIA
	DpsViolationExit(paran);
#endif
	return DPS_OK;
}

int DpsRobotListFree(DPS_ROBOTS *Robots){
	size_t i,j; 
#ifdef WITH_PARANOIA
	void *paran = DpsViolationEnter();
#endif
	
	if(!Robots->nrobots){
#ifdef WITH_PARANOIA
	  DpsViolationExit(paran);
#endif
	  return 0;
	}
	for(i=0;i<Robots->nrobots;i++){
		for(j=0;j<Robots->Robot[i].nrules;j++){
			DPS_FREE(Robots->Robot[i].Rule[j].path);
		}
		DPS_FREE(Robots->Robot[i].hostinfo);
		DPS_FREE(Robots->Robot[i].Rule);
	}
	DPS_FREE(Robots->Robot);
	Robots->nrobots=0;
#ifdef WITH_PARANOIA
	DpsViolationExit(paran);
#endif
	return 0;
}

static DPS_ROBOT_RULE DpsRobotErrRule = {DPS_METHOD_VISITLATER, ""};

DPS_ROBOT_RULE* DpsRobotRuleFind(DPS_AGENT *Indexer, DPS_ROBOTS *Robots, DPS_URL *URL) {
        DPS_ROBOT_RULE *r;
	DPS_ROBOT *robot;
	size_t j;
	DPS_SERVER	*rServer;
	char		*rurl = NULL;
	DPS_DOCUMENT	*rDoc;
	size_t        rurlen;
	int           status, result;
#ifdef WITH_PARANOIA
	void *paran = DpsViolationEnter();
#endif
	if (strcasecmp(DPS_NULL2EMPTY(URL->schema), "http")) { /* robots.txt exist only for http scheme */
	  return NULL;
	}

	rurlen = 32 + strlen(DPS_NULL2EMPTY(URL->schema)) + strlen(DPS_NULL2EMPTY(URL->hostinfo)) +
	  strlen(DPS_NULL2EMPTY(URL->specific));
	rurl = (char*)DpsMalloc(rurlen);
	if ( rurl == NULL) {
	  return &DpsRobotErrRule;
	}

	robot = DpsRobotFind(Robots, DPS_NULL2EMPTY(URL->hostinfo));
	if (robot == NULL) {
	  
	  rDoc = DpsDocInit(NULL);
	  DpsSpiderParamInit(&rDoc->Spider);
	  rDoc->Buf.maxsize = (size_t)DpsVarListFindInt(&Indexer->Conf->Vars, "MaxDocSize", DPS_MAXDOCSIZE);
	  if ((rDoc->Buf.buf = (char*)DpsMalloc(rDoc->Buf.maxsize + 1)) == NULL) {
	    DpsDocFree(rDoc);
	    return &DpsRobotErrRule;
	  }
	  rDoc->Buf.buf[0]='\0';

	  dps_snprintf(rurl, rurlen, "%s://%s/robots.txt", DPS_NULL2EMPTY(URL->schema), DPS_NULL2EMPTY(URL->hostinfo));
	  DpsVarListAddStr(&rDoc->Sections, "URL", rurl);
	  DpsVarListReplaceInt(&rDoc->Sections, "URL_ID", DpsStrHash32(rurl));
	  DpsURLParse(&rDoc->CurURL, rurl);
	  DpsLog(Indexer, DPS_LOG_INFO, "ROBOTS: %s", rurl);

	  rServer = DpsServerFind(Indexer, &Indexer->Conf->Servers, rurl, URL->charset_id, NULL);
	  DpsDocAddDocExtraHeaders(Indexer, rDoc);
	  DpsDocAddConfExtraHeaders(Indexer->Conf, rDoc);

	  if (rServer != NULL) {
	    DpsVarListReplaceLst(&rDoc->Sections, &rServer->Vars, NULL, "*");
	    DpsDocAddServExtraHeaders(rServer, rDoc);
	  }
	  DpsVarListLog(Indexer, &rDoc->RequestHeaders, DPS_LOG_DEBUG, "ROBOTS.Request");
					
	  DpsDocLookupConn(Indexer, rDoc);
	  result = DpsGetURL(Indexer, rDoc);
	  DpsParseHTTPResponse(Indexer, rDoc);
	  DpsDocProcessResponseHeaders(Indexer, rDoc);
	  DpsVarListLog(Indexer, &rDoc->Sections, DPS_LOG_DEBUG, "ROBOTS.Response");

	  if ((status = DpsVarListFindInt(&rDoc->Sections, "Status", 0)) == DPS_HTTP_STATUS_OK) {
					
	    result = 
	      DpsRobotParse(Indexer, rServer, rDoc->Buf.content, (char*)DPS_NULL2EMPTY(rDoc->CurURL.hostinfo));
/*	  } else if ((status == 0) || (status >= 500)) {
	    Doc.method = DPS_METHOD_VISITLATER;*/
	  } else {
	    DpsRobotAddEmpty(&Indexer->Conf->Robots, (char*)DPS_NULL2EMPTY(URL->hostinfo));
	  }

	  DpsDocFree(rDoc);
	  
	  robot = DpsRobotFind(Robots, DPS_NULL2EMPTY(URL->hostinfo));

	}
	if (robot != NULL) {
	  dps_snprintf(rurl, rurlen, "%s%s%s", URL->path, URL->filename, URL->query_string); 
		for(j=0;j<robot->nrules;j++){
			/* FIXME: compare full URL */
			if(!strncmp(rurl /*DPS_NULL2EMPTY(URL->path)*/, robot->Rule[j].path, strlen(robot->Rule[j].path))) {
				if(robot->Rule[j].cmd!=DPS_METHOD_DISALLOW) {
				  DPS_FREE(rurl);
#ifdef WITH_PARANOIA
				  DpsViolationExit(paran);
#endif
				  return NULL;
				} else {
				  r = &robot->Rule[j];
				  DPS_FREE(rurl);
#ifdef WITH_PARANOIA
				  DpsViolationExit(paran);
#endif
				  return r;
				}
			}
		}
	}
	DPS_FREE(rurl);
#ifdef WITH_PARANOIA
	DpsViolationExit(paran);
#endif
	return NULL;
}

int DpsRobotParse(DPS_AGENT *Indexer, DPS_SERVER *Srv, char *content, char *hostinfo) {
        DPS_ENV *Conf = Indexer->Conf;
        DPS_ROBOTS *Robots = &Conf->Robots;
	DPS_ROBOT *robot;
	int rule = 0, common = 0, my = 0;
	char *s,*e,*lt;
	const char *UA = (Srv != NULL) ? DpsVarListFindStr(&Srv->Vars, "Request.User-Agent", DPS_USER_AGENT) :
	  DpsVarListFindStr(&Indexer->Conf->Vars, "Request.User-Agent", DPS_USER_AGENT);

	/* Wipe out any existing (default) rules for this host */
	robot=DeleteRobotRules(Robots, DPS_NULL2EMPTY(hostinfo));
	if (robot == NULL) robot = DpsRobotAddEmpty(Robots, DPS_NULL2EMPTY(hostinfo));
	if(robot==NULL) return(DPS_ERROR);
	
	if(content==NULL) return(DPS_OK);

	s = dps_strtok_r(content, "\r\n", &lt);
	while(s){
		if(*s=='#'){
		}else
		if(!(strncasecmp(s,"User-Agent:",11))){
			char * agent;
			
			agent=DpsTrim(s+11," \t\r\n");
			
			/* The "*" User-Agent is important only */
			/* if no other rules apply */
			if(!strcmp(agent,"*") && (!robot->nrules)){
			  if (!my) {
				rule = 1;
				common = 1;
			  }
			}else
			if(!strncasecmp(agent, UA, strlen(agent))) {
			        rule = 1; my = 1;
				if (common) {
				  robot = DeleteRobotRules(Robots, DPS_NULL2EMPTY(hostinfo));
				  common = 0;
				}
			}else{
				rule = 0;
			}
		}else
		if((!(strncasecmp(s, "Disallow", 8))) && (rule)) {
			if((e=strchr(s+9,'#')))*e=0;
			e=s+9;DPS_SKIP(e," \t");s=e;
			DPS_SKIPN(e," \t");*e=0;
			if(s && *s) {
			  if(AddRobotRule(robot,DPS_METHOD_DISALLOW, s)) {
				  DpsLog(Indexer, DPS_LOG_ERROR, "AddRobotRule error: no memory ?");
					return(DPS_ERROR);
			  }
			} else { /* Empty Disallow == Allow all */
			  if(AddRobotRule(robot, DPS_METHOD_GET, "/")) {
				  DpsLog(Indexer, DPS_LOG_ERROR, "AddRobotRule error: no memory ?");
					return(DPS_ERROR);
			  }
			}
		}else
		if((!(strncasecmp(s, "Allow", 5))) && (rule)) {
			if((e=strchr(s+6,'#')))*e=0;
			e=s+6;DPS_SKIP(e," \t");s=e;
			DPS_SKIPN(e," \t");*e=0;
			if(s && *s){
			  if(AddRobotRule(robot,DPS_METHOD_GET, s)) {
				  DpsLog(Indexer, DPS_LOG_ERROR, "AddRobotRule error: no memory ?");
					return(DPS_ERROR);
			  }
			}
		}
		s = dps_strtok_r(NULL, "\n\r", &lt);
	}
	return(DPS_OK);
}
